@@ -1,4 +1,4 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
+/* vim: set sw=8: -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
  *  gnome-print-pdf.c: the PDF backend
  *
@@ -52,13 +52,6 @@ typedef struct _GnomePrintPdfClass GnomePrintPdfClass;
 typedef struct _GnomePrintPdfPage   GnomePrintPdfPage;
 typedef struct _GnomePrintPdfObject GnomePrintPdfObject;
 
-/* FIXME */
-typedef struct {
-	char *font_name;
-	gint is_basic_14;
-} ps_internal_font;
-/* */
-
 typedef enum {
 	PDF_COMPRESSION_NONE,
 	PDF_COMPRESSION_FLATE,
@@ -179,7 +172,7 @@ static gint gnome_print_pdf_print_bpath (GnomePrintPdf *pdf, const ArtBpath *bpa
 static gchar *gnome_print_pdf_get_date (void);
 
 static gint  gnome_print_pdf_graphic_mode_set (GnomePrintPdf *pdf, PdfGraphicMode mode);
-static gint  gnome_print_pdf_page_fprintf (GnomePrintPdf *pdf, const char *format, ...);
+static gint  gnome_print_pdf_page_fprintf (GnomePrintPdf *pdf, const char *format, ...) G_GNUC_PRINTF (2, 3);
 
 static GnomePrintContextClass *parent_class;
 
@@ -368,7 +361,7 @@ gnome_print_pdf_page_write (GnomePrintPdf *pdf, gchar *text)
 }
 
 
-/* Note "format" should be locale independent, so it should not use %f,
+/* Note "format" should be locale independent, so it should not use %.2f,
  * and pdf does not support scientific notation so it should not use %g
  */
 static gint
@@ -391,11 +384,12 @@ gnome_print_pdf_page_fprintf (GnomePrintPdf *pdf, const char *format, ...)
 	return len;
 }
 
-/* Allowed conversion specifiers are 'e', 'E', 'f', 'F', 'g' and 'G'.
+/* Allowed conversion specifiers are 'f', 'F'
  *
  * NOTE: From pdf ref 1.5 v 6 : section 3.2.2
  * 	PDF does not support the PostScript syntax for numbers with nondecimal
  * 	radices (such as 16#FFFE) or in exponential format (such as 6.02E23).
+ * 	so %g and %e will not work
  */
 static gint
 gnome_print_pdf_page_print_double (GnomePrintPdf *pdf,
@@ -582,7 +576,7 @@ gnome_print_pdf_image (GnomePrintContext *pc, const gdouble *ctm, const guchar *
 {
 	GnomePrintPdfPage *page;
 	GnomePrintPdf *pdf;
-	gint hex_size, row, object_num, ret = 0, mask_object_num;
+	gint hex_size, row, object_num, ret = 0, mask_object_num = 0;
 	guchar *hex;
 	gint length_object_num, len = 0;
 
@@ -669,22 +663,22 @@ gnome_print_pdf_image (GnomePrintContext *pc, const gdouble *ctm, const guchar *
 		page->used_color_images = TRUE;
 
 	gnome_print_pdf_page_write (pdf, "q" EOL);
-	gnome_print_pdf_page_print_double (pdf, "%f",
+	gnome_print_pdf_page_print_double (pdf, "%.2f",
 				      ctm[0]);
 	gnome_print_pdf_page_write (pdf, " ");
-	gnome_print_pdf_page_print_double (pdf, "%f",
+	gnome_print_pdf_page_print_double (pdf, "%.2f",
 				      ctm[1]);
 	gnome_print_pdf_page_write (pdf, " ");
-	gnome_print_pdf_page_print_double (pdf, "%f",
+	gnome_print_pdf_page_print_double (pdf, "%.2f",
 				      ctm[2]);
 	gnome_print_pdf_page_write (pdf, " ");
-	gnome_print_pdf_page_print_double (pdf, "%f",
+	gnome_print_pdf_page_print_double (pdf, "%.2f",
 				      ctm[3]);
 	gnome_print_pdf_page_write (pdf, " ");
-	gnome_print_pdf_page_print_double (pdf, "%f",
+	gnome_print_pdf_page_print_double (pdf, "%.2f",
 				      ctm[4]);
 	gnome_print_pdf_page_write (pdf, " ");
-	gnome_print_pdf_page_print_double (pdf, "%f",
+	gnome_print_pdf_page_print_double (pdf, "%.2f",
 				      ctm[5]);
 	gnome_print_pdf_page_write (pdf, " cm" EOL);
 	gnome_print_pdf_page_fprintf (pdf, "/Im%d Do" EOL, object_num);
@@ -723,16 +717,20 @@ gnome_print_pdf_assign_code_to_glyph (GnomePrintPdf *pdf, gint glyph)
 static gint
 gnome_print_pdf_glyphlist (GnomePrintContext *pc, const gdouble *ctm, GnomeGlyphList *gl)
 {
-	GnomePrintPdfPage *page;
 	static gdouble id[] = {1.0, 0.0, 0.0, 1.0, 0.0, 0.0};
-	GnomePrintPdf *pdf;
-	GnomePosGlyphList *pgl;
-	gint ret, s;
-	gdouble dx, dy;
-	gdouble Tm [6], scale [6];
-	ArtPoint point;
 
-	pdf = GNOME_PRINT_PDF (pc);
+	GnomePrintPdf	  *pdf = GNOME_PRINT_PDF (pc);
+	GnomePrintPdfPage *page;
+	GnomePosGlyphList *pgl;
+	GnomePosGlyph const *glyph;
+	ArtPoint advance;
+	char	 octal [ 1 /* header char  \\ */ +
+			11 /* 32 bit int in octal */ +
+			 1 /* trailing null */];
+	gint	 ret, s, master_glyph, glyph_page, code;
+	gint	 x_adjust, y_adjust, cur_y_adjust = 0;
+	gdouble	 Tm [6], scale [6], base_y, prev_x = 0;
+	enum { OUTSIDE, IN_ARRAY, IN_GLYPHS } state;
 
 	gnome_print_pdf_graphic_mode_set (pdf, PDF_GRAPHIC_MODE_TEXT);
 
@@ -741,9 +739,6 @@ gnome_print_pdf_glyphlist (GnomePrintContext *pc, const gdouble *ctm, GnomeGlyph
 
 	pgl = gnome_pgl_from_gl (gl, id, GNOME_PGL_RENDER_DEFAULT);
 
-	dx = ctm[4];
-	dy = ctm[5];
-
 	for (s = 0; s < pgl->num_strings; s++) {
 		GnomePosString *ps;
 		GnomeFont *font;
@@ -752,12 +747,6 @@ gnome_print_pdf_glyphlist (GnomePrintContext *pc, const gdouble *ctm, GnomeGlyph
 		ps = pgl->strings + s;
 		font = gnome_rfont_get_font (ps->rfont);
 
-		/* Currently only truetype subsetting supported */
-
-		if (!NEEDED_SUBSETTING(font)) {
-			gnome_print_pdf_set_font_real (pdf, font, FALSE, 0);
-		}
-
 		ret = gnome_print_pdf_set_color_real (pdf,
 						      PDF_COLOR_GROUP_FILL,
 						      ((ps->color >> 24) & 0xff) / 255.0,
@@ -765,113 +754,95 @@ gnome_print_pdf_glyphlist (GnomePrintContext *pc, const gdouble *ctm, GnomeGlyph
 						      ((ps->color >>  8) & 0xff) / 255.0);
 
 		art_affine_scale (scale, font->size, font->size);
+		scale[4]	  = pgl->glyphs[ps->start].x;
+		scale[5] = base_y = pgl->glyphs[ps->start].y;
 		art_affine_multiply (Tm, scale, ctm);
 
-		point.x = pgl->glyphs[ps->start].x;
-		point.y = pgl->glyphs[ps->start].y;
-		art_affine_point (&point, &point, ctm);;
-
-		gnome_print_pdf_page_print_double (pdf, "%f", Tm [0]);
+		gnome_print_pdf_page_print_double (pdf, "%.2f", Tm [0]);
 		gnome_print_pdf_page_write (pdf, " ");
-		gnome_print_pdf_page_print_double (pdf, "%f", Tm [1]);
+		gnome_print_pdf_page_print_double (pdf, "%.2f", Tm [1]);
 		gnome_print_pdf_page_write (pdf, " ");
-		gnome_print_pdf_page_print_double (pdf, "%f", Tm [2]);
+		gnome_print_pdf_page_print_double (pdf, "%.2f", Tm [2]);
 		gnome_print_pdf_page_write (pdf, " ");
-		gnome_print_pdf_page_print_double (pdf, "%f", Tm [3]);
+		gnome_print_pdf_page_print_double (pdf, "%.2f", Tm [3]);
 		gnome_print_pdf_page_write (pdf, " ");
-		gnome_print_pdf_page_print_double (pdf, "%f", dx + point.x);
+		gnome_print_pdf_page_print_double (pdf, "%.2f", Tm [4]);
 		gnome_print_pdf_page_write (pdf, " ");
-		gnome_print_pdf_page_print_double (pdf, "%f", dy + point.y);
+		gnome_print_pdf_page_print_double (pdf, "%.2f", Tm [5]);
 		gnome_print_pdf_page_write (pdf, " Tm" EOL);
 
 		/* Build string */
-		if (!NEEDED_SUBSETTING(font)) {
-			ret = gnome_print_pdf_page_write (pdf, "(");
-			for (i = ps->start; i < ps->start + ps->length; i++) {
-				gint glyph;
-				glyph = pgl->glyphs[i].glyph;
-				glyph = gnome_print_pdf_assign_code_to_glyph (pdf, glyph);
-				ret = gnome_print_pdf_page_fprintf (pdf, "\\%o", glyph);
-				g_return_val_if_fail (ret >= 0, ret);
-			}
-		} else {
-			for (i = ps->start; i < ps->start + ps->length; i++) {
+		state = OUTSIDE;
+		for (i = ps->start; i < ps->start + ps->length; i++) {
+			glyph	     = pgl->glyphs + i;
+			master_glyph = glyph->glyph;
 
-				/* Glyph position in parent font */
-
-				gint master_glyph;
-
-				/* Glyph in subfont */
-
-				gint glyph;
-
-				/* Pages of 256 glyphs, including starting */
-				/* null glyph				   */
-
-				gint page;
-
-				master_glyph = (gint)(pgl->glyphs[i].glyph);
-				page  = master_glyph / 255;
-
-				/* Initially */
-
-				if (i==ps->start) {
-					current_page = page;
-					gnome_print_pdf_set_font_real (pdf, font, TRUE, page);
-					ret = gnome_print_pdf_page_write (pdf, "(");
-
-				/* There's a ttf subfont to be set*/
-
-				} else if (current_page != page) {
-					current_page = page;
-					(void)gnome_print_pdf_page_write (pdf, ") Tj" EOL);
-					gnome_print_pdf_set_font_real (pdf, font, TRUE, page);
-					ret = gnome_print_pdf_page_write (pdf, "(");
+			if (i != ps->start) {
+				state = IN_GLYPHS;
+				x_adjust = (int)  (advance.x + .5) -
+					   (int) ((1000. *(glyph->x - prev_x) / font->size) + .5);
+				y_adjust = (int)  (advance.y + .5) -
+					   (int) ((1000. *(glyph->y - base_y) / font->size) + .5);
+#if 0
+				fprintf (stderr, "advance [%d].x = %4.0f\tvs a delta of\t%g\n", i,
+					 advance.x, 1000. *(glyph->x - prev_x) / font->scale);
+				fprintf (stderr, "advance [%d].y = %4.0f\tvs a delta of\t%g\n", i,
+					 advance.y, 1000. *(glyph->y - base_y) / font->scale);
+#endif
+				if (x_adjust != 0) {
+					gnome_print_pdf_page_fprintf (pdf, ") %d" EOL, x_adjust);
+					state = IN_ARRAY;
 				}
 
+				if (cur_y_adjust != y_adjust) {
+					if (state == IN_GLYPHS)
+						gnome_print_pdf_page_write (pdf, ")");
+					gnome_print_pdf_page_fprintf (pdf, " %d] TJ " EOL,
+						(cur_y_adjust = y_adjust));
+					state = OUTSIDE;
+				}
+				if (NEEDED_SUBSETTING (font)) {
+					glyph_page = master_glyph / 255;
+					if (glyph_page != current_page) {
+						if (state == IN_GLYPHS)
+							gnome_print_pdf_page_write (pdf, ")] TJ" EOL);
+						else if (state == IN_ARRAY)
+							gnome_print_pdf_page_write (pdf, "] TJ" EOL);
+						state = OUTSIDE;
+						gnome_print_pdf_set_font_real (pdf, font, TRUE,
+							(current_page = glyph_page));
+					}
+				}
+			} else if (NEEDED_SUBSETTING (font)) {
+				current_page   = master_glyph / 255;
+				gnome_print_pdf_set_font_real (pdf, font, TRUE, current_page);
+			} else
+				gnome_print_pdf_set_font_real (pdf, font, FALSE, 0);
+
+			if (NEEDED_SUBSETTING (font)) {
 				gnome_font_face_pso_mark_glyph (pdf->selected_font->pso, master_glyph);
+				/* For each sub ttf font the first glyph is
+				 * null and 255th one is a valid glyph, also 1
+				 * needed to be added to each subglyph  for
+				 * taking into account first null glyph added
+				 * to every subfont */
+				code = master_glyph ? master_glyph % 255 + 1 : master_glyph;
+			} else
+				code = gnome_print_pdf_assign_code_to_glyph (pdf, master_glyph);
 
-				/* For each sub ttf font the first glyph is  */
-				/* null and 255th one is a valid glyph, also */
-				/* 1 needed to be added to each subglyph     */
-				/* for taking into account first null glyph  */
-				/* added to every subfont	             */
-
-				glyph = master_glyph ? master_glyph % 255 + 1 : master_glyph;
-
-				ret = gnome_print_pdf_page_fprintf (pdf, "\\%o", glyph);
-				g_return_val_if_fail (ret >= 0, ret);
+			if (state < IN_GLYPHS) {
+				ret = gnome_print_pdf_page_write (pdf, (state == OUTSIDE) ? "[(" : "(");
+				state = IN_GLYPHS;
 			}
-		}
 
-		ret = gnome_print_pdf_page_write (pdf, ") Tj" EOL);
+			g_snprintf (octal, sizeof(octal), "\\%o", code);
+			ret = gnome_print_pdf_page_write (pdf, octal);
+			g_return_val_if_fail (ret >= 0, ret);
 
-/* FIXME: We are not honoring xy movements of a glyphlist right now. But rather
- * just assuming that the glyphlist isn't any different than then character
- * spacing that the font metrics provide. */
-#if 0
-		g_return_val_if_fail (ret >= 0, ret);
-		/* Build array */
-		ret = gnome_print_pdf_page_write (pdf, "[");
-		g_return_val_if_fail (ret >= 0, ret);
-		for (i = ps->start + 1; i < ps->start + ps->length; i++) {
-			ret = gnome_print_pdf_page_print_double (pdf, "%f",
-				     pgl->glyphs[i].x - pgl->glyphs[i-1].x);
-			g_return_val_if_fail (ret >= 0, ret);
-			ret = gnome_print_pdf_page_write (pdf, " ");
-			g_return_val_if_fail (ret >= 0, ret);
-			ret = gnome_print_pdf_page_print_double (pdf, "%f",
-				     pgl->glyphs[i].y - pgl->glyphs[i-1].y);
-			g_return_val_if_fail (ret >= 0, ret);
-			ret = gnome_print_pdf_page_write (pdf, " ");
-			g_return_val_if_fail (ret >= 0, ret);
+			gnome_font_face_get_glyph_stdadvance (font->face, master_glyph, &advance);
+			prev_x = glyph->x;
 		}
-		ret = gnome_print_pdf_page_write (pdf, "0 0] ");
-		g_return_val_if_fail (ret >= 0, ret);
-		/* xyshow */
-		ret = gnome_print_pdf_page_write (pdf, "xyshow" EOL);
-		g_return_val_if_fail (ret >= 0, ret);
-#endif
+		ret = gnome_print_pdf_page_write (pdf, ")] TJ" EOL);
 	}
 
 	gnome_pgl_destroy (pgl);
@@ -1252,7 +1223,7 @@ gnome_print_pdf_write_fonts (GnomePrintPdf *pdf)
 	list = pdf->fonts;
 	while (list) {
 		font = list->data;
-		gnome_print_pdf_font_print_widths   (pdf, font);
+		gnome_print_pdf_font_print_widths (pdf, font);
 		/* '/Encoding' not needed when the font is symbolic */
 		if (!NEEDED_SUBSETTING(font))
 			gnome_print_pdf_font_print_encoding (pdf, font);
@@ -1374,29 +1345,64 @@ gnome_print_pdf_close_write_last_objects (GnomePrintPdf *pdf)
 	return ret;
 }
 
+static int 
+subsetfontname_cmp (gconstpointer a, gconstpointer b)
+{
+	GnomeFontFace *face_a = ((GnomePrintPdfFont *) a)->face;
+	GnomeFontFace *face_b = ((GnomePrintPdfFont *) b)->face;
+
+	return strcmp((const char *)face_a->psname, (const char *)face_b->psname);
+}
+
+#define UCS2_UPPER_BOUND 0xFFFF
+ 
 static void
 gnome_print_embed_all_pdf_fonts (GnomePrintPdf *pdf)
 {
-	GList *list;
+	GList *list, *prev_list = NULL, *copy_list;
+	gint *glyph2unicode;
 
 	g_return_if_fail (pdf != NULL);
+	
+	glyph2unicode = (gint *) calloc (UCS2_UPPER_BOUND, sizeof(gint));
+
+	copy_list = list = g_list_copy ((GList *)pdf->fonts);
+
+	list = g_list_sort (list, subsetfontname_cmp);
 
-	list = pdf->fonts;
 	while (list) {
+		gint font_change = 1;
 		GnomePrintPdfFont *font;
+
 		font = list->data;
-		gnome_print_embed_pdf_font (pdf, font);
+
+		if (prev_list)
+			font_change = subsetfontname_cmp (prev_list->data, list->data);
+		if (font_change && font && font->face && font->face->ft_face) {
+			int i;
+			for(i = 0; i < UCS2_UPPER_BOUND; i++) {
+				int glyph = FT_Get_Char_Index (font->face->ft_face, i);
+				glyph2unicode[glyph] = i;
+			}
+		}
+		gnome_print_embed_pdf_font (pdf, font, glyph2unicode);
+		prev_list = list;
 		list = list->next;
 	}
+
+	g_list_free (copy_list);
+
+	free (glyph2unicode);
 }
 
+#undef UCS2_UPPER_BOUND
 
 static gint
 gnome_print_pdf_close (GnomePrintContext *pc)
 {
 	GnomePrintPdfPage *page;
 	GnomePrintPdf *pdf;
-	gint ret;
+	gint ret = 0;
 
 	pdf = GNOME_PRINT_PDF (pc);
 
@@ -1446,13 +1452,12 @@ gnome_print_pdf_set_line (GnomePrintPdf *pdf)
 
 	if (gp_gc_get_line_flag (ctx->gc) == GP_GC_FLAG_CLEAR)
 		return GNOME_PRINT_OK;
-	gnome_print_pdf_page_print_double (pdf, "%f",
+	gnome_print_pdf_page_print_double (pdf, "%.2f",
 					   gp_gc_get_linewidth (ctx->gc));
 	gnome_print_pdf_page_fprintf (pdf, " w %d J %d j ",
 				      gp_gc_get_linecap (ctx->gc),
-				      gp_gc_get_linejoin (ctx->gc),
-				      gp_gc_get_miterlimit (ctx->gc));
-	gnome_print_pdf_page_print_double (pdf, "%f",
+				      gp_gc_get_linejoin (ctx->gc));
+	gnome_print_pdf_page_print_double (pdf, "%.2f",
 					   gp_gc_get_miterlimit (ctx->gc));
 	gnome_print_pdf_page_write (pdf, " M" EOL);
 	gp_gc_set_line_flag (ctx->gc, GP_GC_FLAG_CLEAR);
@@ -1476,10 +1481,10 @@ gnome_print_pdf_set_dash (GnomePrintPdf *pdf)
 	gnome_print_pdf_page_write (pdf, "[");
 	for (i = 0; i < dash->n_dash; i++) {
 		gnome_print_pdf_page_write (pdf, " ");
-		gnome_print_pdf_page_print_double (pdf, "%f", dash->dash[i]);
+		gnome_print_pdf_page_print_double (pdf, "%.2f", dash->dash[i]);
 	}
 	gnome_print_pdf_page_write (pdf, "]");
-	gnome_print_pdf_page_print_double (pdf, "%f",
+	gnome_print_pdf_page_print_double (pdf, "%.2f",
 				 dash->n_dash > 0 ? dash->offset : 0.0);
 	gnome_print_pdf_page_write (pdf, " d" EOL);
 	gp_gc_set_dash_flag (ctx->gc, GP_GC_FLAG_CLEAR);
@@ -1571,8 +1576,8 @@ gnome_print_pdf_font_print_widths (GnomePrintPdf *pdf, GnomePrintPdfFont *font)
 			gnome_print_pdf_fprintf (pdf, EOL);
 			col = 0;
 		}
-		col += gnome_print_pdf_print_double (pdf, "%f", point.x);
-		col += gnome_print_pdf_fprintf (pdf, " ");
+		col += gnome_print_pdf_print_double (pdf, "%.1f", point.x);
+		gnome_print_pdf_print_sized (pdf, " ", 1);
 		i++;
 	}
 
@@ -1734,9 +1739,9 @@ gnome_print_pdf_font_print_descriptor (GnomePrintPdf *pdf, GnomePrintPdfFont *fo
 	g_object_get (G_OBJECT (face), "XHeight", &val, NULL);
 	xheight = (gint) val;
 	g_object_get (G_OBJECT (face), "FontBBox", &rect, NULL);
-	g_free (rect);
 	art_drect_copy (&new_rect, rect);
 	my_fix_drect (&new_rect);
+	g_free (rect);
 
 	object_number_descriptor = gnome_print_pdf_object_new (pdf);
 	*object_number_descriptor_ret = object_number_descriptor;
@@ -1751,13 +1756,13 @@ gnome_print_pdf_font_print_descriptor (GnomePrintPdf *pdf, GnomePrintPdfFont *fo
 				 "/FontBBox [",
 				 ascent, capheight,
 				 descent * -1, flags);
-	gnome_print_pdf_print_double (pdf, "%f", new_rect.x0);
+	gnome_print_pdf_print_double (pdf, "%.2f", new_rect.x0);
 	gnome_print_pdf_fprintf (pdf, " ");
-	gnome_print_pdf_print_double (pdf, "%f", new_rect.y0);
+	gnome_print_pdf_print_double (pdf, "%.2f", new_rect.y0);
 	gnome_print_pdf_fprintf (pdf, " ");
-	gnome_print_pdf_print_double (pdf, "%f", new_rect.x1);
+	gnome_print_pdf_print_double (pdf, "%.2f", new_rect.x1);
 	gnome_print_pdf_fprintf (pdf, " ");
-	gnome_print_pdf_print_double (pdf, "%f", new_rect.y1);
+	gnome_print_pdf_print_double (pdf, "%.2f", new_rect.y1);
 	gnome_print_pdf_fprintf (pdf,
 				 "]" EOL
 				 "/FontName /%s" EOL
@@ -1787,7 +1792,8 @@ gnome_print_pdf_font_print_descriptor (GnomePrintPdf *pdf, GnomePrintPdfFont *fo
  */
 
 static guchar *
-gnome_print_pdf_get_subfont_name (guchar *basefont, guint unique_subfont_id) {
+gnome_print_pdf_get_subfont_name (guchar *basefont, guint unique_subfont_id)
+{
 	int i;
 	guchar *full_basefont;
 
@@ -1807,8 +1813,10 @@ gnome_print_pdf_get_subfont_name (guchar *basefont, guint unique_subfont_id) {
 }
 
 void
-gnome_print_embed_pdf_font (GnomePrintPdf *pdf, GnomePrintPdfFont *font) {
+gnome_print_embed_pdf_font (GnomePrintPdf *pdf, GnomePrintPdfFont *font, gint *glyph2unicode)
+{
 	gint object_number_descriptor = 0;
+	gint object_number_tounicode = 0;
 	guchar *basefont_name;
 	gboolean plan_b = FALSE; /* TRUE if we could not embed the font */
 
@@ -1821,7 +1829,11 @@ gnome_print_embed_pdf_font (GnomePrintPdf *pdf, GnomePrintPdfFont *font) {
 				   gnome_font_face_get_ps_name (font->face));
 			plan_b = TRUE;
 			font->is_type_1 = TRUE; /* It isn't, but the substitute is */
-		}
+		} 
+
+		if (NEEDED_SUBSETTING(font))
+			gnome_print_pdf_font_tounicode (pdf, font, glyph2unicode, &object_number_tounicode);
+		
 	}
 
 	basefont_name = gnome_print_pdf_get_subfont_name ((guchar *)gnome_font_face_get_ps_name (font->face), font->object_number);
@@ -1856,6 +1868,12 @@ gnome_print_embed_pdf_font (GnomePrintPdf *pdf, GnomePrintPdfFont *font) {
 					  object_number_descriptor);
 	}
 
+	if (NEEDED_SUBSETTING(font)) {
+		gnome_print_pdf_fprintf (pdf, 
+					"/ToUnicode %d 0 R" EOL, 
+					object_number_tounicode);
+	}
+
 	gnome_print_pdf_object_end   (pdf, font->object_number, FALSE);
 }
 
@@ -1994,9 +2012,9 @@ gnome_print_pdf_print_bpath (GnomePrintPdf *pdf, const ArtBpath *bpath)
 				gnome_print_pdf_page_write (pdf, "h" EOL);
 			closed = FALSE;
 			started = FALSE;
-			gnome_print_pdf_page_print_double (pdf, "%f", bpath->x3);
+			gnome_print_pdf_page_print_double (pdf, "%.2f", bpath->x3);
 			gnome_print_pdf_page_write (pdf, " ");
-			gnome_print_pdf_page_print_double (pdf, "%f", bpath->y3);
+			gnome_print_pdf_page_print_double (pdf, "%.2f", bpath->y3);
 			gnome_print_pdf_page_write (pdf, " m" EOL);
 			break;
 		case ART_MOVETO:
@@ -2004,29 +2022,29 @@ gnome_print_pdf_print_bpath (GnomePrintPdf *pdf, const ArtBpath *bpath)
 				gnome_print_pdf_page_write (pdf, "h" EOL);
 			closed = TRUE;
 			started = TRUE;
-			gnome_print_pdf_page_print_double (pdf, "%f", bpath->x3);
+			gnome_print_pdf_page_print_double (pdf, "%.2f", bpath->x3);
 			gnome_print_pdf_page_write (pdf, " ");
-			gnome_print_pdf_page_print_double (pdf, "%f", bpath->y3);
+			gnome_print_pdf_page_print_double (pdf, "%.2f", bpath->y3);
 			gnome_print_pdf_page_write (pdf, " m" EOL);
 			break;
 		case ART_LINETO:
-			gnome_print_pdf_page_print_double (pdf, "%f", bpath->x3);
+			gnome_print_pdf_page_print_double (pdf, "%.2f", bpath->x3);
 			gnome_print_pdf_page_write (pdf, " ");
-			gnome_print_pdf_page_print_double (pdf, "%f", bpath->y3);
+			gnome_print_pdf_page_print_double (pdf, "%.2f", bpath->y3);
 			gnome_print_pdf_page_write (pdf, " l" EOL);
 			break;
 		case ART_CURVETO:
-			gnome_print_pdf_page_print_double (pdf, "%f", bpath->x1);
+			gnome_print_pdf_page_print_double (pdf, "%.2f", bpath->x1);
 			gnome_print_pdf_page_write (pdf, " ");
-			gnome_print_pdf_page_print_double (pdf, "%f", bpath->y1);
+			gnome_print_pdf_page_print_double (pdf, "%.2f", bpath->y1);
 			gnome_print_pdf_page_write (pdf, " ");
-			gnome_print_pdf_page_print_double (pdf, "%f", bpath->x2);
+			gnome_print_pdf_page_print_double (pdf, "%.2f", bpath->x2);
 			gnome_print_pdf_page_write (pdf, " ");
-			gnome_print_pdf_page_print_double (pdf, "%f", bpath->y2);
+			gnome_print_pdf_page_print_double (pdf, "%.2f", bpath->y2);
 			gnome_print_pdf_page_write (pdf, " ");
-			gnome_print_pdf_page_print_double (pdf, "%f", bpath->x3);
+			gnome_print_pdf_page_print_double (pdf, "%.2f", bpath->x3);
 			gnome_print_pdf_page_write (pdf, " ");
-			gnome_print_pdf_page_print_double (pdf, "%f", bpath->y3);
+			gnome_print_pdf_page_print_double (pdf, "%.2f", bpath->y3);
 			gnome_print_pdf_page_write (pdf, " c" EOL);
 			break;
 		default:
@@ -2096,20 +2114,23 @@ gnome_print_pdf_fprintf (GnomePrintPdf *pdf, const char *format, ...)
 	return len;
 }
 
+/* Allowed conversion specifiers are 'f', 'F'
+ *
+ * NOTE: From pdf ref 1.5 v 6 : section 3.2.2
+ * 	PDF does not support the PostScript syntax for numbers with nondecimal
+ * 	radices (such as 16#FFFE) or in exponential format (such as 6.02E23).
+ * 	so %g and %e will not work
+ */
 gint
 gnome_print_pdf_print_double (GnomePrintPdf *pdf, const gchar *format, gdouble x)
 {
- 	gchar *text;
+ 	gchar text [G_ASCII_DTOSTR_BUF_SIZE];
 	gint len;
 
-	text = g_new (gchar, G_ASCII_DTOSTR_BUF_SIZE);
 	g_ascii_formatd (text, G_ASCII_DTOSTR_BUF_SIZE, format, x);
-
 	len = strlen (text);
 	gnome_print_transport_write (GNOME_PRINT_CONTEXT (pdf)->transport, text, len);
 	pdf->offset += len;
-	g_free (text);
-
 	return len;
 }
 
@@ -2119,7 +2140,6 @@ gnome_print_pdf_print_sized (GnomePrintPdf *pdf, const char *content, gint len)
 {
 	gnome_print_transport_write (GNOME_PRINT_CONTEXT (pdf)->transport, content, len);
 	pdf->offset += len;
-
 	return len;
 }
 
