@@ -204,7 +204,7 @@ gpa_printer_new_from_tree (xmlNodePtr tree)
 	xmlNodePtr node;
 	GPANode *settings, *printer, *model;
 	xmlChar *name, *id, *version;
-	const char **syslangs = NULL;
+	const char* const *syslangs = NULL;
 	
 	g_return_val_if_fail (tree != NULL, NULL);
 	g_return_val_if_fail (tree->name != NULL, NULL);
