@@ -90,27 +90,6 @@ static xmlChar *model_xml_template =
 "  </Options>"
 "</Model>";
 
-static GPANode *
-gpa_model_new_from_xml (char *string)
-{
-	GPANode *model;
-	xmlNodePtr root;
-	xmlDocPtr doc;
-	
-	doc = xmlParseDoc ((unsigned char *)string);
-	if (!doc) {
-		g_warning ("Could not parse model xml");
-		return NULL;
-	}
-	
-	root = doc->xmlRootNode;
-	model = gpa_model_new_from_tree (root);
-	
-	xmlFreeDoc (doc);
-	
-	return model;
-}
-
 static GPANode *
 load_paper_sizes (papi_attribute_t **attrs, GPANode *parent)
 {
@@ -504,7 +483,7 @@ get_model (const char *name, papi_service_t service)
 	
 	if (model == NULL) {
 		xml = g_strdup_printf ((char *)model_xml_template, id, nickname);
-		model = gpa_model_new_from_xml (xml);
+		model = gpa_model_new_from_xml_str (xml);
 		g_free (xml);
 
 		output = gpa_node_lookup (model, (guchar *)"Options.Output");
@@ -707,6 +686,11 @@ all_not_found:
 			return;
 		}
 
+		if (!printers) {
+			papiServiceDestroy (service);
+			return;
+		}
+
 		for (i = 0; printers[i] != NULL; i++)
 			;
 
