@@ -25,7 +25,9 @@
 
 #include "sft.h"
 #include "gnome-print-tt-subset.h"
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif
 
 static char *
 gnome_print_tmpnam (void)
